import os
import re
import subprocess
import sys
from subprocess import Popen

__all__ = ["CommandError", "run"]

from textwrap import dedent

from typing import Optional, List, Union

from envo.misc import is_linux, is_windows
from envo import console


class CommandError(RuntimeError):
    pass


def _run(command: str,
         ignore_errors=False,
         print_output=True,
         verbose: Optional[bool]=None,
         background=False) -> Optional[str]:

    if verbose is None:
        verbose = os.environ.get("ENVO_VERBOSE_RUN", False)

    command_extra = re.sub(r"\\(?:\t| )*\n(?:\t| )*", "", command)

    if is_windows():
        command_extra = command_extra.strip()
        command_extra = command_extra.replace("\r", "")
        command_extra = command_extra.replace("\n", " & ")
        popen_cmd = ["cmd.exe", "/c", command_extra]
    elif is_linux():
        options = ["set -uoe pipefail", "shopt -s globstar"]
        command_extra = "\n".join(options) + "\n" + command_extra
        popen_cmd = ["/bin/bash", "--rcfile", "/dev/null", "-c", command_extra]
    else:
        raise NotImplementedError()

    if verbose and print_output:
        dedent_cmd = dedent(command.strip())
        console.rule(f"[bold rgb(225,221,0)]{dedent_cmd}", align="center", style="rgb(255,0,255)")

    proc = Popen(popen_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=os.environ)

    if background:
        return None

    buffer = []

    while True:
        c = proc.stdout.read(1)
        if not c or c == b"\xf0":
            break

        if print_output:
            try:
                sys.stdout.buffer.write(c)
                sys.stdout.flush()
            except ValueError:
                break
        buffer.append(c)

    ret_code = proc.wait()

    if ret_code != 0 and not ignore_errors:
        sys.exit(ret_code)
    else:
        ret = b"".join(buffer)
        ret = ret.decode("utf-8")
        return ret


def run(
        command: Union[str, List[str]],
        ignore_errors=False,
        print_output=True,
        verbose: Optional[bool]=None,
        background=False,
        progress_bar: Optional[str] = None
) -> List[Optional[str]]:
    # join multilines
    ret = []
    if not isinstance(command, list):
        command = [command]

    if progress_bar is not None:
        from rich.progress import track
        for c in track(command, description=progress_bar):
            ret.append(_run(c, ignore_errors, print_output, verbose, background))
    else:
        for c in command:
            ret.append(_run(c, ignore_errors, print_output, verbose, background))

    return ret



#!/usr/bin/env python3
import hashlib
import os
import sys
from collections import OrderedDict
from dataclasses import dataclass
from pathlib import Path
from typing import Any, ClassVar, Dict, List, Optional, Type

import envo.e2e
from envo import const, logger, logging, misc, shell
from envo.env import ShellEnv, Env
from envo.misc import Callback, EnvoError, FilesWatcher, import_env_from_file
from envo.shell import FancyShell, PromptBase, PromptState, Shell
from envo.status import Status

package_root = Path(os.path.realpath(__file__)).parent
templates_dir = package_root / "templates"


__all__ = ["_main"]

DEFAULT_STAGE = "Default"


class CantFindEnvFile(EnvoError):
    def __init__(self):
        super().__init__("Couldn't find any env!\n" 'Forgot to run envo init" first?')


class NormalPrompt(PromptBase):
    msg: str = ""

    @property
    def p_name(self) -> str:
        return f"({self.name})" if self.name else ""

    @property
    def p_msg(self) -> str:
        return f"{{BOLD_RED}}{self.msg}{{RESET}}\n" if self.msg else ""

    def __init__(self) -> None:
        super().__init__()

        self.state_prefix_map = {
            PromptState.LOADING: lambda: f"{self.p_msg}{const.emojis['loading']}{self.p_name}{self.default}",
            PromptState.NORMAL: lambda: f"{self.p_msg}{self.emoji}{self.p_name}{self.default}",
        }


class HeadlessMode:
    @dataclass
    class Links:
        shell: Shell

    @dataclass
    class Sets:
        stage: str
        restart_nr: int
        msg: str
        env_path: Path

    @dataclass
    class Callbacks:
        restart: Callback
        on_error: Callback

    shell_env: ShellEnv
    reloader_enabled: bool = False
    blocking: bool = True

    def __init__(self, se: Sets, li: Links, calls: Callbacks) -> None:
        self.se = se
        self.li = li
        self.calls = calls

        self.extra_watchers = []

        self.status = Status(
            calls=Status.Callbacks(
                on_ready=Callback(self._on_ready),
                on_not_ready=Callback(self._on_not_ready),
            )
        )

        self.shell_env = None

        self.li.shell.set_fulll_traceback_enabled(True)

        logger.set_level(logging.Level.INFO)

        logger.debug("Creating Headless Mode")

    def _on_ready(self) -> None:
        self.prompt.state = PromptState.NORMAL
        self.li.shell.set_prompt(self.prompt.as_str())

    def _on_not_ready(self) -> None:
        self.prompt.state = PromptState.LOADING
        self.li.shell.set_prompt(self.prompt.as_str())

    def unload(self) -> None:
        if self.shell_env:
            self.shell_env._unload()

        self.li.shell.calls.reset()

    def init(self) -> None:
        self.li.shell.set_context({"logger": logger})

        self._create_env()

        self.prompt = NormalPrompt()
        self.prompt.state = PromptState.LOADING
        self.prompt.emoji = self.shell_env.env.meta.emoji
        self.prompt.name = self.shell_env.get_name()
        self.prompt.msg = self.se.msg

        self.li.shell.set_prompt(str(self.prompt))

        self.li.shell.set_variable("env", self.shell_env)
        self.li.shell.set_variable("environ", os.environ)

        self.shell_env.validate()
        self.shell_env.activate()

        self.shell_env.load()

    def get_env_file(self) -> Path:
        return self.se.env_path

    def _create_env_object(self, file: Path) -> ShellEnv:
        env = import_env_from_file(file).ThisEnv()

        shell_env = ShellEnv(
            li=ShellEnv._Links(shell=self.li.shell, status=self.status, env=env),
            calls=ShellEnv._Callbacks(
                restart=self.calls.restart,
                on_error=self.calls.on_error,
            ),
            se=ShellEnv._Sets(
                reloader_enabled=self.reloader_enabled,
                blocking=self.blocking,
                extra_watchers=self.extra_watchers,
            ),
        )
        return shell_env

    def _create_env(self) -> None:
        env_file = self.get_env_file()
        logger.debug(f'Creating Env from file "{env_file}"')

        # unload modules
        for m in list(sys.modules.keys())[:]:
            if m.startswith("env_"):
                sys.modules.pop(m)
        try:
            self.shell_env = self._create_env_object(env_file)
        except ImportError as exc:
            logger.traceback()
            raise EnvoError(f"""Couldn't import "{env_file}" ({exc}).""")


class NormalMode(HeadlessMode):
    @dataclass
    class Links(HeadlessMode.Links):
        pass

    @dataclass
    class Sets(HeadlessMode.Sets):
        pass

    @dataclass
    class Callbacks(HeadlessMode.Callbacks):
        pass

    shell_env: ShellEnv
    reloader_enabled: bool = True
    blocking: bool = False

    def __init__(self, se: Sets, li: Links, calls: Callbacks) -> None:
        super(NormalMode, self).__init__(se=se, li=li, calls=calls)
        self.se = se
        self.li = li
        self.calls = calls

        self.li.shell.set_fulll_traceback_enabled(False)

        logger.set_level(logging.Level.ERROR)
        logger.debug("Creating NormalMode")

    def stop(self) -> None:
        self.shell_env._exit()


class EmergencyMode(HeadlessMode):
    @dataclass
    class Links(HeadlessMode.Links):
        previous_env: ShellEnv

    @dataclass
    class Sets(HeadlessMode.Sets):
        pass

    @dataclass
    class Callbacks(HeadlessMode.Callbacks):
        pass

    reloader_enabled: bool = True
    blocking: bool = False

    def __init__(self, se: Sets, li: Links, calls: Callbacks) -> None:
        super().__init__(se=se, li=li, calls=calls)

        self.se = se
        self.li = li
        self.calls = calls

        logger.set_level(logging.Level.ERROR)
        logger.debug("Creating EmergencyMode")
        self.li.shell.set_fulll_traceback_enabled(False)

        self.li.shell.calls.on_exit = Callback(self.stop)

    def get_env_file(self) -> Path:
        self.extra_watchers = []

        if self.li.previous_env:
            for w in self.li.previous_env.reloader.env_watchers:
                self.extra_watchers.append(
                FilesWatcher(
                    FilesWatcher.Sets(
                        root=w.root,
                        include=["env_*.py"],
                        exclude=[],
                    ),
                    calls=FilesWatcher.Callbacks(on_event=Callback(None)),
                ))

        # remove duplicates
        self.extra_watchers = list({obj.se.root: obj for obj in self.extra_watchers}.values())
        return Path(__file__).parent / "emergency_env.py"

    def stop(self) -> None:
        self.shell_env._exit()


class EnvoBase:
    @dataclass
    class Sets:
        stage: str

    shell: shell.Shell
    mode: Optional[HeadlessMode]
    env_dirs: List[Path]

    def __init__(self, se: Sets):
        self.se = se
        logger.set_level(logging.Level.INFO)
        self.mode = None

        self.env_dirs = self._get_env_dirs()

        self.restart_count = -1

    def _get_env_dirs(self) -> List[Path]:
        ret = []
        path = Path(".").absolute()
        while True:
            if path.parent == path:
                break

            for p in path.glob("env_*.py"):
                if p.parent not in ret:
                    ret.append(p.parent)

            path = path.parent

        return ret

    def find_env(self) -> Path:
        # TODO: Test this
        if self.se.stage != DEFAULT_STAGE:
            matches: Dict[Path, const.Stage] = OrderedDict()
            for d in self.env_dirs:
                for p in d.glob("env_*.py"):
                    stage = const.STAGES.filename_to_stage(p.name)
                    if stage:
                        matches[p] = stage

            results = [p for p, s in matches.items() if self.se.stage == s.name]
            if results:
                return results[0]
        else:
            for d in self.env_dirs:
                matches: Dict[Path, const.Stage] = OrderedDict()
                for p in d.glob("env_*.py"):
                    stage = const.STAGES.filename_to_stage(p.name)
                    if stage:
                        matches[p] = stage

                results = sorted(
                    matches.items(), key=lambda x: x[1].priority, reverse=True
                )
                if results:
                    return results[0][0]

        raise CantFindEnvFile()

    @property
    def data_dir_name(self) -> str:
        hash_object = hashlib.md5(str(self.find_env()).encode("utf-8"))
        ret = hash_object.hexdigest()
        return ret

    def init(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError()

    def restart(self) -> None:
        self.init()

    def single_command(self, command: str) -> None:
        raise NotImplementedError()

    def dry_run(self) -> None:
        raise NotImplementedError()

    def dump(self) -> None:
        raise NotImplementedError()


class EnvoHeadless(EnvoBase):
    @dataclass
    class Sets(EnvoBase.Sets):
        stage: str

    shell: shell.Shell
    mode: HeadlessMode

    def __init__(self, se: Sets):
        super().__init__(se)
        self.se = se
        logger.set_level(logging.Level.INFO)

        if not self.env_dirs:
            raise CantFindEnvFile

        self.restart_count = -1

    def on_error(self) -> None:
        pass

    def init(self, *args: Any, **kwargs: Any) -> None:
        self.restart_count += 1
        self.shell.reset()

        self.mode = HeadlessMode(
            se=HeadlessMode.Sets(
                stage=self.se.stage,
                restart_nr=self.restart_count,
                msg="",
                env_path=self.find_env(),
            ),
            calls=HeadlessMode.Callbacks(
                restart=Callback(self.restart), on_error=Callback(self.on_error)
            ),
            li=HeadlessMode.Links(shell=self.shell),
        )
        self.mode.init()

    def single_command(self, command: str) -> None:
        self.shell = Shell.create(Shell.Callbacks(), data_dir_name=self.data_dir_name)
        self.init()

        try:
            self.shell.default(command)
        except SystemExit as e:
            sys.exit(e.code)
        else:
            sys.exit(self.shell.last_return_code)
        finally:
            self.mode.unload()

    def dry_run(self) -> None:
        self.shell = Shell.create(Shell.Callbacks(), data_dir_name=self.data_dir_name)
        self.init()
        content = "\n".join(
            [f'export {k}="{v}"' for k, v in self.mode.shell_env.env.get_env_vars().items()]
        )
        print(content)

    def dump(self) -> None:
        self.shell = Shell.create(Shell.Callbacks(), data_dir_name=self.data_dir_name)
        self.init()
        path = self.mode.shell_env.env.dump_dot_env()
        logger.info(f"Saved envs to {str(path)} 💾", print_msg=True)


class Envo(EnvoBase):
    @dataclass
    class Sets(EnvoBase.Sets):
        pass

    environ_before = Dict[str, str]
    inotify: FilesWatcher
    quit: bool
    mode: HeadlessMode

    def __init__(self, se: Sets) -> None:
        super().__init__(se)

        self.quit: bool = False
        self.environ_before = os.environ.copy()  # type: ignore
        logger.set_level(logging.Level.ERROR)

    def init(self, *args: Any, **kwargs: Any) -> None:
        self.restart_count += 1
        try:
            self.shell.reset()

            if self.mode:
                self.mode.unload()

            self.mode = NormalMode(
                se=NormalMode.Sets(
                    stage=self.se.stage,
                    restart_nr=self.restart_count,
                    msg="",
                    env_path=self.find_env(),
                ),
                li=NormalMode.Links(shell=self.shell),
                calls=NormalMode.Callbacks(
                    restart=Callback(self.restart), on_error=Callback(self.on_error)
                ),
            )
            self.mode.init()

        except BaseException as exc:
            self.on_error(exc)

    def on_error(self, exc: BaseException) -> None:
        msg = misc.get_envo_relevant_traceback(exc)
        msg = "".join(msg)
        msg = msg.rstrip()

        logger.error(msg)

        if self.mode:
            self.mode.unload()

        self.mode = EmergencyMode(
            se=EmergencyMode.Sets(
                stage=self.se.stage,
                restart_nr=self.restart_count,
                msg=msg,
                env_path=self.find_env(),
            ),
            li=EmergencyMode.Links(shell=self.shell, previous_env=self.mode.shell_env),
            calls=EmergencyMode.Callbacks(
                restart=Callback(self.restart), on_error=Callback(None)
            ),
        )

        self.mode.init()

    def spawn_shell(self) -> None:
        """
        :param type: shell type
        """

        def on_ready():
            pass

        self.shell = FancyShell.create(
            calls=FancyShell.Callbacks(on_ready=Callback(on_ready)),
            data_dir_name=self.data_dir_name,
        )
        self.init()

        self.mode.shell_env.on_shell_create()

        self.shell.start()
        self.mode.unload()


class EnvoCreator:
    @dataclass
    class Sets:
        stage: str

    def __init__(self, se: Sets) -> None:
        logger.debug("Starting EnvoCreator")
        self.se = se

    def _create_from_templ(self, stage: str) -> None:
        """
        Create env file from template.

        :param templ_file:
        :param output_file:
        :param is_comm:
        :return:
        """
        output_file = Path(f"env_{stage}.py")

        parent = "env_comm" if stage != "comm" else None

        if output_file.exists():
            raise EnvoError(f"{str(output_file)} file already exists.")

        env_dir = Path(".").absolute()
        package_name = misc.dir_name_to_pkg_name(env_dir.name)
        class_name = (
            f"{misc.dir_name_to_class_name(package_name)}{stage.capitalize()}Env"
        )

        context = {
            "class_name": class_name,
            "name": env_dir.name,
            "stage": stage,
            "emoji": const.STAGES.get_stage_name_to_emoji().get(stage, "🙂"),
            "parent_import": f"\nfrom {parent} import ThisEnv as ParentEnv\n" if parent else "",
            "base_class": "ParentEnv" if parent else Env.__name__,
            "this_env": const.THIS_ENV
        }

        templ_file = Path("env.py.templ")
        misc.render_py_file(
            templates_dir / templ_file, output=output_file, context=context
        )

    def create(self) -> None:
        if not self.se.stage:
            self.se.stage = "comm"

        self._create_from_templ(self.se.stage)

        if self.se.stage != "comm" and not Path("env_comm.py").exists():
            self._create_from_templ("comm")

        print(f"Created {self.se.stage} environment 🍰!")


@dataclass
class BaseOption:
    stage: str
    flesh: str

    keywords: ClassVar[str] = NotImplemented

    def run(self) -> None:
        raise NotImplementedError()


@dataclass
class Command(BaseOption):
    def run(self) -> None:
        envo.e2e.envo = env_headless = EnvoHeadless(EnvoHeadless.Sets(stage=self.stage))
        env_headless.single_command(self.flesh)


@dataclass
class DryRun(BaseOption):
    def run(self) -> None:
        envo.e2e.envo = env_headless = EnvoHeadless(EnvoHeadless.Sets(stage=self.stage))
        env_headless.dry_run()


@dataclass
class Dump(BaseOption):
    def run(self) -> None:
        envo.e2e.envo = env_headless = EnvoHeadless(EnvoHeadless.Sets(stage=self.stage))
        env_headless.dump()


@dataclass
class Version(BaseOption):
    def run(self) -> None:
        from envo.__version__ import __version__

        print(__version__)


@dataclass
class Init(BaseOption):
    def run(self) -> None:
        stage = "comm" if self.stage == DEFAULT_STAGE else self.stage
        envo_creator = EnvoCreator(EnvoCreator.Sets(stage=stage))
        envo_creator.create()


@dataclass
class Start(BaseOption):
    def run(self) -> None:
        envo.e2e.envo = e = Envo(Envo.Sets(stage=self.stage))
        e.spawn_shell()


option_name_to_option: Dict[str, Type[BaseOption]] = {
    "-c": Command,
    "run": Command,
    "dry-run": DryRun,
    "dump": Dump,
    "": Start,
    "init": Init,
    "version": Version,
}


def _main() -> None:
    logger.debug("Starting")

    argv = sys.argv[1:]
    keywords = ["init", "dry-run", "version", "dump", "run"]

    stage = DEFAULT_STAGE
    if argv and argv[0] not in keywords:
        stage = argv[0]
        option_name = argv[1] if len(argv) >= 2 else ""
        flesh = " ".join(argv[2:])
    else:
        option_name = argv[0] if len(argv) >= 1 else ""
        flesh = " ".join(argv[1:])

    option = option_name_to_option[option_name](stage, flesh=flesh)

    try:
        option.run()
    except EnvoError as e:
        logger.error(str(e), print_msg=True)
        if envo.e2e.enabled:
            envo.e2e.on_exit()
        sys.exit(1)
    finally:
        if envo.e2e.enabled:
            envo.e2e.on_exit()


if __name__ == "__main__":
    _main()



import inspect
import re
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, Optional

__all__ = ["STAGES", "emojis"]

DEFAULT_EMOJI = "🙂"
THIS_ENV = "ThisEnv"


@dataclass
class Stage:
    name: str
    priority: int  # higher - activated first when no stage is provided
    emoji: str


class STAGES:
    COMM = Stage("comm", 90, "👌")
    LOCAL = Stage("local", 100, "🐣")
    TEST = Stage("test", 80, "🛠")
    CI = Stage("ci", 70, "🧪")
    STAGE = Stage("stage", 60, "🤖")
    PROD = Stage("prod", 50, "🔥")

    @classmethod
    def get_all_stages(cls) -> Dict[str, Stage]:
        ret = {}
        for _, obj in inspect.getmembers(cls):
            if isinstance(obj, Stage):
                ret[obj.name] = obj

        return ret

    @classmethod
    def get_stage_name_to_emoji(cls) -> Dict[str, str]:
        stages = cls.get_all_stages()

        ret = defaultdict(lambda: DEFAULT_EMOJI)
        for s in stages.values():
            ret[s.name] = s.emoji

        return ret

    @classmethod
    def filename_to_stage(cls, filename: str) -> Optional[Stage]:
        stages = cls.get_all_stages()
        matches_groups = re.search(r"env_(.*)\.py", filename)
        if not matches_groups:
            raise RuntimeError(f"Not an envo file ({filename})")

        matches = matches_groups.groups()
        stage_name = matches[0]

        stage = stages.get(stage_name, None)

        if not stage:
            stage = Stage(stage_name, priority=10, emoji=DEFAULT_EMOJI)

        return stage


emojis: Dict[str, str] = {"loading": "⏳", "emergency": "❌"}



import re
import sys
import traceback
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional

import loguru
from loguru._colorizer import Colorizer
from pygments.styles import get_style_by_name
from rhei import Stopwatch
from xonsh.pyghooks import XonshConsoleLexer


class Level(Enum):
    DEBUG = 0
    INFO = 1
    WARNING = 2
    ERROR = 3


@dataclass
class Msg:
    level: Level
    body: str
    time: float  # s
    descriptor: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.body = str(self.body).lstrip()

    def print(self) -> None:
        print(self.render_all(with_color=True))

    def _fix_formatting(self, text: str) -> str:
        from loguru._colorizer import AnsiParser

        parser = AnsiParser()

        for match in parser._regex_tag.finditer(text):
            markup, tag = match.group(0), match.group(1)

            if tag not in {"lvl", "level"}:
                ansi = parser._get_ansicode(tag)
                if ansi is None:
                    text = text.replace(
                        markup, markup.replace("<", "< ").replace(">", " >")
                    )

        return text

    def render_message(self, with_color: bool = True) -> str:
        def color(clr: str) -> str:
            if with_color:
                return clr
            else:
                return ""

        from pygments import highlight
        from pygments.formatters.terminal import TerminalFormatter

        metadata = ""
        if self.metadata:
            metadata = str(self.metadata)

        descriptor = ""
        if self.descriptor:
            descriptor = (
                f"{color('<green>')}({str(self.descriptor)}) {color('</green>')}"
            )
        msg = f"@{self.time:.4f}]{descriptor}{self.body}; {metadata}"

        if with_color:
            # fix escaping colors
            msg = self._fix_formatting(msg)
            msg = highlight(
                msg,
                XonshConsoleLexer(),
                TerminalFormatter(style=get_style_by_name("emacs")),
            )
            msg = Colorizer.ansify(msg)
        return msg

    def render_all(self, with_color: bool = True) -> str:
        return f"[{self.level.name:<5}{self.render_message(with_color=with_color)}"

    def __repr__(self) -> str:
        return self.render_all(with_color=False)


@dataclass
class MsgFilter:
    level: Optional[Level] = None
    body_re: Optional[str] = None
    time_later: Optional[float] = None
    time_before: Optional[float] = None
    metadata_re: Optional[Dict[str, Any]] = None

    # matchers
    def matches_level(self, msg: Msg) -> bool:
        return self.level is None or msg.level == self.level

    def matches_body(self, msg: Msg) -> bool:
        return self.body_re is None or bool(re.match(self.body_re, msg.body, re.DOTALL))

    def matches_time_later(self, msg: Msg) -> bool:
        return self.body_re is None or msg.time >= self.time_later

    def matches_time_before(self, msg: Msg) -> bool:
        return self.body_re is None or msg.time < self.time_before

    def matches_metadata(self, msg: Msg) -> bool:
        if not self.metadata_re:
            return True

        for k, v in self.metadata_re.items():
            msg_value = msg.metadata.get(k)
            if msg_value is None:
                return False

            if not re.match(v, msg_value, re.DOTALL):
                return False

        return True

    def matches_all(self, msg: Msg) -> bool:
        return (
            self.matches_level(msg)
            and self.matches_body(msg)
            and self.matches_time_later(msg)
            and self.matches_time_before(msg)
            and self.matches_metadata(msg)
        )


class Messages(list):
    content = List[Msg]

    def print(self) -> None:
        ret = []

        for m in self:
            m.print()

        ret = "\n".join(ret)

        print(ret)


class Logger:
    messages: Messages
    level: Level
    parent: Optional["Logger"]
    descriptor: Optional[str]
    name: str

    def __init__(
        self,
        name: str,
        parent: Optional["Logger"] = None,
        descriptor: Optional[str] = None,
    ) -> None:
        self.name = name
        self.parent = parent
        self.descriptor = descriptor

        self.messages = Messages()
        self.level = Level.INFO

        self.set_level(Level.INFO)

        self.sw = Stopwatch()
        self.sw.start()

        loguru.logger.remove()
        loguru.logger.add(
            sys.stdout,
            format="<blue>{message}</blue>",
            level="DEBUG",
            filter=lambda x: x["level"].name == "DEBUG",
        )
        loguru.logger.add(
            sys.stdout,
            format="<bold>{message}</bold>",
            level="INFO",
            filter=lambda x: x["level"].name == "INFO",
        )
        loguru.logger.add(
            sys.stderr,
            format="<bold><yellow>{message}</yellow></bold>",
            level="WARNING",
            filter=lambda x: x["level"].name == "WARNING",
        )
        loguru.logger.add(
            sys.stderr,
            format="<bold><red>{message}</red></bold>",
            level="ERROR",
            filter=lambda x: x["level"].name == "ERROR",
        )

    def create_child(self, name: str, descriptor: str) -> "Logger":
        logger = Logger(parent=self, name=name, descriptor=descriptor)
        logger.sw = self.sw
        return logger

    def clean(self) -> None:
        self.messages = Messages()

    def set_level(self, level: Level) -> None:
        self.level = level

    def _log(self, msg: Msg) -> None:
        self.messages.append(msg)

    def log(
        self,
        message: str,
        level: Level,
        metadata: Optional[Dict[str, Any]] = None,
        print_msg=False,
    ) -> None:
        msg = Msg(
            level,
            message,
            self.sw.value,
            metadata=metadata or {},
            descriptor=self.descriptor,
        )
        if print_msg:
            loguru.logger.log(level.name, message)

        self._log(msg)

        if self.parent:
            self.parent._log(msg)

    def debug(
        self, message: str, metadata: Optional[Dict[str, Any]] = None, print_msg=False
    ) -> None:
        self.log(message, Level.DEBUG, metadata, print_msg)

    def info(
        self, message: str, metadata: Optional[Dict[str, Any]] = None, print_msg=False
    ) -> None:
        self.log(message, Level.INFO, metadata, print_msg)

    def warning(
        self, message: str, metadata: Optional[Dict[str, Any]] = None, print_msg=False
    ) -> None:
        self.log(message, Level.WARNING, metadata, print_msg)

    def error(
        self, message: str, metadata: Optional[Dict[str, Any]] = None, print_msg=False
    ) -> None:
        self.log(message, Level.ERROR, metadata, print_msg)

    #
    # def get_user_code_exception(root: Path) -> str:
    #     lines = traceback.format_exception(*sys.exc_info())
    #     users_code = False
    #     ret = []
    #     for l in lines:
    #         if str(root) in l:
    #             users_code = True
    #         if not users_code:
    #             continue
    #
    #         if l.startswith("  "):
    #             l = l[2:]
    #
    #         ret.append(l)
    #
    #     return "".join(ret)

    def traceback(self) -> None:
        lines = traceback.format_exception(*sys.exc_info())
        text = "".join(lines).strip()
        self.log(text, level=Level.ERROR, print_msg=True)

    def get_msgs(self, filter: MsgFilter) -> List[Msg]:
        filtered: List[Msg] = []
        for m in self.messages:
            if filter.matches_all(m):
                filtered.append(m)

        return filtered

    def print_all(self) -> None:
        for m in self.messages:
            m.print()

    def tail(self, messages_n: int) -> None:
        for m in self.messages[-messages_n:]:
            m.print()

    def save(self, file: Path) -> None:
        pass


logger = Logger(name="root")



from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple  # noqa: F401

import envo  # noqa: F401
from envo import (  # noqa: F401
    Plugin,
    VirtualEnv,
    boot_code,
    command,
    const,
    context,
    logger,
    oncreate,
    ondestroy,
    onload,
    onstderr,
    onstdout,
    onunload,
    postcmd,
    precmd,
    run,
    Env
)


class EmergencyEnv(Env):  # type: ignore
    class Meta(Env.Meta):  # type: ignore
        root = Path(__file__).parent.absolute()
        stage: str = "emergency"
        emoji: str = const.emojis["emergency"]
        plugins: List[Plugin] = []
        name: str = "emergency"
        version: str = "0.1.0"
        watch_files: List[str] = []
        ignore_files: List[str] = []

    class Environ(Env.Environ):
        ...

    e: Environ

    def init(self) -> None:
        pass


ThisEnv = EmergencyEnv
